/* Copyright (c) (2017,2019-2023) Apple Inc. All rights reserved.
 *
 * corecrypto is licensed under Apple Inc.’s Internal Use License Agreement (which
 * is contained in the License.txt file distributed with corecrypto) and only to
 * people who accept that license. IMPORTANT:  Any license rights granted to you by
 * Apple Inc. (if any) are limited to internal use within your organization only on
 * devices and computers you own or control, for the sole purpose of verifying the
 * security characteristics and correct functioning of the Apple Software.  You may
 * not, directly or indirectly, redistribute the Apple Software or any portions thereof.
 */

#include "cc_debug.h"
#include "ccrsa_internal.h"
#include <corecrypto/ccsha2.h>
#include "ccrng_zero.h"

#include "fipspost.h"
#include "fipspost_priv.h"
#include "fipspost_post_rsa_sig.h"

#define RSA_NBITS 2048

static const uint8_t fipspost_post_rsa_test_key[] = "\x30\x82\x04\xa1\x02\x01\x00\x02\x82\x01\x01\x00\x88\xdd\x23\x6b\xe4\x29\x46\xff\xe9\x25\x5e\x07\x9d\x39\xe2\xc9\x00\x50\x93\xeb\x32\x88\x88\xd3\xd8\xba\x3a\x1b\xc8\x6c\x61\x3f\xe2\xe0\x9f\x8c\xbc\x75\x3a\x51\xdf\xec\x6b\xab\xc6\x0b\xca\xe6\x61\x7d\xba\xb5\xc6\x06\x53\x27\x83\xa1\xe1\xf4\x15\x5a\xc0\x4f\x4e\xb2\xee\x03\x0a\x2b\x43\x8a\x1f\xac\x14\x54\xa1\x21\xf1\xfe\x9a\x88\x68\x02\xd7\xb1\xca\xba\xb9\xd4\x24\x11\xc2\x89\x70\x80\xfa\x75\x24\x47\x92\x52\x40\x04\x47\x7d\x9a\x60\xdc\xa9\xd7\x2c\xe6\x8c\x88\x71\xf3\x69\x0e\x59\xa7\xd9\x51\xa7\xa8\x2d\x93\x2f\xe3\x70\xd5\x64\x9e\x69\x06\x94\xb0\x11\xa2\x0d\x64\x41\x9a\x44\x32\xd8\xa8\x72\x35\x52\x7d\xc8\xf9\xa7\x03\x3f\x12\x77\x22\xb2\xe0\xbf\x0a\x07\xb6\x5b\xe9\x13\x24\x60\xf7\xc4\xb1\x4d\x1e\xe1\xa7\x48\xdb\x1d\xcc\xe3\xef\x04\x1c\xac\xb0\x98\xdc\x68\x6a\xb5\x6f\x02\x72\x8e\xec\xe3\x4a\xba\x79\xab\x80\x3b\xb7\x18\xf4\xfc\x9d\x33\xd5\xa1\x38\xd4\x7c\xa7\xfa\x39\x45\xc3\x86\x75\x1f\x0b\x93\x2d\x50\x2b\xc5\x33\x60\xde\x40\xd8\x1f\xa4\x2a\xc8\x32\xb4\x5f\xbb\x1c\xab\x67\xcf\xf7\xe3\xaa\x8d\x54\x85\xea\xc1\x0c\xdd\x02\x01\x03\x02\x82\x01\x00\x16\xcf\x85\xe7\x50\xb1\x8b\xd5\x51\x86\x3a\x56\x9a\x34\x50\x76\xd5\x62\xc3\x51\xdd\xc1\x6c\x23\x4e\xc9\xb4\x59\xf6\xbc\xba\xdf\xfb\x25\x6f\xec\xca\x13\x89\xb8\x4f\xfc\xbc\x9c\xa1\x01\xf7\x26\x65\x94\xf4\x73\xa1\x01\x0d\xdb\xeb\x45\xa5\xa8\xae\x39\xca\xb7\xe2\x73\x27\xab\x2c\x5c\x8b\x41\xaf\xf2\x03\x63\x70\x30\x52\xff\xc4\x6c\x11\x55\xce\x9d\xa1\xc9\xc9\xa3\x5b\x58\x4b\x16\xe8\x15\x7f\x13\x86\x0b\xed\xb8\x60\x00\xb6\x94\xef\x10\x24\xc6\xf9\x32\x26\x6c\xc1\x68\x53\x3c\x2d\x0e\xf1\x4e\xe2\xf1\x46\xb2\x43\x32\x67\x80\xb5\x97\x12\xf7\x44\x58\xea\x78\x45\x89\x51\xb4\xb5\xa5\xc4\x00\xf7\xfa\x77\xbf\x8d\x1f\xa5\xc8\x38\xcf\x03\xeb\x7a\x88\x62\x25\x5d\xf5\x67\xb3\x1e\x5a\x9a\x33\x16\x7c\x8c\x04\xb5\x87\x1f\x39\xbe\x73\x0c\x8b\x14\xf4\x4d\xda\x62\xc8\x31\x66\x57\x10\x20\xaf\xc3\xc5\x73\x86\x7b\xd9\x4d\x93\xe7\x15\x13\x7d\xbd\xa2\xb5\x82\x74\xe5\x05\xb9\x3e\x44\x6c\x93\x43\xc8\xc1\x6d\xc4\xea\x16\x9a\x24\xcd\x14\xcb\xb6\xc0\xb7\xbd\x82\x90\xd7\x86\xb2\x25\x3b\x8a\xe8\x50\xa8\x26\xcd\x55\xd6\xd4\x24\x5e\x56\x2d\xf3\xa5\x02\x81\x81\x00\xbe\x8f\xc2\x63\x3f\x1d\x1e\x7d\xb1\xd0\x2d\x17\xc9\x16\xf6\x33\x54\x16\x4d\x5e\x7b\xe9\x40\x86\x45\x07\xb2\x0c\x5d\xff\x46\x1a\x16\x6c\x7f\x37\xac\x24\x25\x18\xcd\xb2\xea\xb9\x57\x07\x4d\xad\x64\x6c\xbf\xb3\x84\xbb\x04\x89\x49\xcf\xdd\x05\x09\x39\xe4\x61\xdf\x34\xb7\xd7\xf3\xdc\x58\x6d\xfb\x9a\xd4\x13\x02\xe3\x23\xeb\x68\x22\x50\xe8\x2b\xf2\xda\x9a\x64\x91\x59\x48\x15\x91\x64\x40\x21\xd1\xb3\x6b\x9c\xae\x87\x03\x86\x47\xab\xd4\xa1\xae\x05\xbc\x9f\x7f\x2a\xd0\x66\x4a\x3f\x6f\x63\xd2\x93\x18\x96\x14\x4f\xfb\x02\x81\x81\x00\xb7\xdc\xd1\x76\xed\x80\x4e\x01\x7f\x6f\xd3\xbd\xb0\xee\x62\x2e\x46\xbc\x8b\x34\xea\xeb\xee\x84\xd1\xed\xfc\x58\x9c\xf2\xfd\x66\x7d\x72\x57\x0f\x9c\x05\x0d\xda\xb9\x7b\x86\x20\x12\x29\x90\x09\x87\x81\xa4\xb7\xfc\xe6\x6c\xc0\xff\xbe\x82\xe2\xaa\xc8\x7b\xf2\xcb\xaf\x24\x16\x43\xe0\x0b\x34\xac\x99\x41\xaa\x3f\x43\x5f\x40\xf4\x02\xc7\x5a\xea\x8a\x2c\x73\x0a\x34\x55\xc6\xe8\x51\x1d\x4e\xe9\xbe\xbf\xf1\xab\xbe\x91\x56\x6c\x1f\x64\x6a\x7b\xf2\x00\x18\x5a\xfa\x7f\xf7\x10\x9c\xe8\x71\x3d\xc1\xe7\x37\x4f\x99\x07\x07\x02\x81\x80\x7f\x0a\x81\x97\x7f\x68\xbe\xfe\x76\x8a\xc8\xba\x86\x0f\x4e\xcc\xe2\xb9\x88\xe9\xa7\xf0\xd5\xae\xd8\xaf\xcc\x08\x3e\xaa\x2e\xbc\x0e\xf2\xff\x7a\x72\xc2\xc3\x65\xde\x77\x47\x26\x3a\x04\xde\x73\x98\x48\x7f\xcd\x03\x27\x58\x5b\x86\x8a\x93\x58\xb0\xd1\x42\xeb\xea\x23\x25\x3a\xa2\x92\xe5\x9e\xa7\xbc\x8d\x62\x01\xec\xc2\x9c\xf0\x16\xe0\x9a\xc7\xf7\x3c\x66\xed\xb6\x3b\x85\x63\xb6\x42\xd5\x6b\xe1\x22\x47\xbd\xc9\xaf\x57\xae\xda\x72\x8d\xc1\x1e\xae\x7d\xbf\xaa\x1c\x8a\xee\xdc\x2a\x4a\x42\x8c\x62\x10\x64\x0d\x8a\xa7\x02\x81\x80\x7a\x93\x36\x4f\x49\x00\x34\x00\xff\x9f\xe2\x7e\x75\xf4\x41\x74\x2f\x28\x5c\xcd\xf1\xf2\x9f\x03\x36\x9e\xa8\x3b\x13\x4c\xa8\xee\xfe\x4c\x3a\x0a\x68\x03\x5e\x91\xd0\xfd\x04\x15\x61\x71\x0a\xb1\x05\x01\x18\x7a\xa8\x99\x9d\xd5\xff\xd4\x57\x41\xc7\x30\x52\xa1\xdd\x1f\x6d\x64\x2d\x40\x07\x78\x73\x10\xd6\x71\x7f\x82\x3f\x80\xa2\xac\x84\xe7\x47\x06\xc8\x4c\xb1\x78\x39\x2f\x45\x8b\x68\xdf\x46\x7f\x2a\xa1\x1d\x29\xb6\x39\x9d\x6a\x42\xf1\xa7\xf6\xaa\xba\xe7\x51\xaa\xa4\xb5\xbd\xf0\x4b\x7e\x81\x44\xcf\x8a\x66\x04\xaf\x02\x81\x81\x00\xbd\xa5\xef\xf8\x36\x8a\x70\x98\x1e\x80\xde\xad\xbb\x67\x8c\xc7\x60\x62\xc5\x2f\x93\x39\x2c\xf6\x43\x4e\x18\x58\x60\x64\xed\x5b\x4d\xee\x77\xb2\x0c\xad\x3e\xff\x95\xb1\x05\x54\x28\xce\x9b\x0f\x83\xa0\x7d\x13\x7e\xfb\x67\x70\xda\x3c\x1e\xc3\xfb\x1f\x29\xba\x08\xff\x26\xb8\x73\xf4\x9c\x3e\x8e\x06\xc8\xcf\x44\x52\x96\x27\x73\x65\x20\x34\x20\x67\xdc\xca\x16\x36\x3b\x01\x0d\x62\x9a\x7f\x6c\x88\x29\x8a\x5f\x7a\x2e\xd6\xe8\x37\x2f\x3a\xe5\x8a\xa3\x9a\x83\x38\x4a\x62\x58\x3e\xcb\xfe\x0a\x25\x06\x2b\x7c\xa2\x0d\x6f";

static const size_t fipspost_post_rsa_test_key_nbytes = sizeof(fipspost_post_rsa_test_key) - 1;

const unsigned char MESSAGE[] =
    "\x70\x99\x2c\x9d\x95\xa4\x90\x8d\x2a\x94\xb3\xab\x9f\xa1\xcd\x64"
    "\x3f\x12\x0e\x32\x6f\x9d\x78\x08\xaf\x50\xca\xc4\x2c\x4b\x0b\x4e"
    "\xeb\x7f\x0d\x4d\xf3\x03\xa5\x68\xfb\xfb\x82\xb0\xf5\x83\x00\xd2"
    "\x53\x57\x64\x57\x21\xbb\x71\x86\x1c\xaf\x81\xb2\x7a\x56\x08\x2c"
    "\x80\xa1\x46\x49\x9f\xb4\xea\xb5\xbd\xe4\x49\x3f\x5d\x00\xf1\xa4"
    "\x37\xbb\xc3\x60\xdf\xcd\x80\x56\xfe\x6b\xe1\x0e\x60\x8a\xdb\x30"
    "\xb6\xc2\xf7\x65\x24\x28\xb8\xd3\x2d\x36\x29\x45\x98\x2a\x46\x58"
    "\x5d\x21\x02\xef\x79\x95\xa8\xba\x6e\x8a\xd8\xfd\x16\xbd\x7a\xe8"
    "\xf5\x3c\x3d\x7f\xcf\xba\x29\x0b\x57\xce\x7f\x8f\x09\xc8\x28\xd6"
    "\xf2\xd3\xce\x56\xf1\x31\xbd\x94\x61\xe5\x66\x7e\x5b\x73\xed\xac"
    "\x77\xf5\x04\xda\xc4\xf2\x02\xa9\x57\x0e\xb4\x51\x5b\x2b\xf5\x16"
    "\x40\x7d\xb8\x31\x51\x8d\xb8\xa2\x08\x3e\xc7\x01\xe8\xfd\x38\x7c"
    "\x43\x0b\xb1\xa7\x2d\xec\xa5\xb4\x9d\x42\x9c\xf9\xde\xb0\x9c\xc4"
    "\x51\x8d\xc5\xf5\x7c\x08\x9a\xa2\xd3\x42\x0e\x56\x7e\x73\x21\x02"
    "\xc2\xc9\x2b\x88\xa0\x7c\x69\xd7\x09\x17\x14\x0a\xb3\x82\x3c\x63"
    "\xf3\x12\xd3\xf1\x1f\xa8\x7b\xa2\x9d\xa3\xc7\x22\x4b\x4f\xb4\xbc";

#define SIGNATURE POST_FIPS_RESULT_STR(                                \
    "\x5e\xb1\xe4\xb3\x4e\xda\x48\x52\x70\x59\x57\xe4\x22\xe1\xaf\x21" \
    "\xf0\x2d\x77\xe3\x73\xc9\xda\xdb\x69\x83\x31\x0d\x12\xe9\xbf\x40" \
    "\x4c\x19\x37\xa0\xe7\x9d\x18\x5e\xfb\x08\x48\x58\x14\x4d\x0e\x7e" \
    "\xcf\x6f\x7a\x53\x57\x71\xcd\x26\xd8\xee\xb2\xba\x6f\x71\xbc\xf0" \
    "\x16\x80\xe1\x04\xa5\xf0\x4a\x74\x0b\xa4\xee\x5c\x9e\x3f\xef\x75" \
    "\x1f\xcb\x87\x7c\xc2\x3d\xaa\x22\x48\xeb\xf1\x5f\xfb\x27\x88\x59" \
    "\x00\x27\xb9\x5b\x31\x07\x0a\x79\x50\x8d\xe8\x68\x9d\xa6\x33\x67" \
    "\x17\xc7\x94\xff\xb6\xcb\x82\x6d\xf0\x72\x0a\xb2\x05\x78\x9f\xdd" \
    "\x84\xfe\xdd\xb9\xb9\x2c\xeb\x49\xcf\x24\xf6\x06\x57\x9c\x6e\xe7" \
    "\x57\xde\x02\xd4\x2e\xc8\xfe\x48\xd0\xa7\xfb\xff\x18\xcc\x5a\x4e" \
    "\x7e\xec\x17\x88\x9a\xcf\x94\x39\xb6\xdc\xe7\x48\xa5\x4c\xba\x24" \
    "\xb8\xb5\xcd\x04\xbc\xb8\xb4\x36\xe1\x9c\x42\x33\x78\xb0\xcc\x98" \
    "\xd4\x9b\xc8\xb3\x53\x8e\xf1\xa3\x1b\x43\xe4\x81\x5c\x82\x13\x01" \
    "\xdc\xc3\x13\x51\x09\xd4\x1f\xa8\xb4\x2e\x26\x6c\x8e\x35\x62\x64" \
    "\x2f\x10\x35\xef\x5f\x65\x40\x61\xfc\x4e\x52\x05\xe7\x38\xba\x92" \
    "\x9e\x5f\xf0\xe3\x76\x1f\x36\xba\x42\x2d\xdb\x7b\xb6\xb1\x5c\x0b")

const size_t MESSAGE_LEN = sizeof(MESSAGE) - 1;

CC_INLINE int fipspost_post_rsa_sign(uint32_t fips_mode)
{
    /* Sign / Verify */
    cc_size n = ccn_nof(RSA_NBITS);
    ccrsa_full_ctx_decl_nbits(RSA_NBITS, full_key);
    ccrsa_ctx_n(full_key)=n;

    // Import the key
    if (0 != ccrsa_import_priv(full_key, fipspost_post_rsa_test_key_nbytes, fipspost_post_rsa_test_key))
    {
        failf("import");
        return CCPOST_GENERIC_FAILURE;
    }

    uint8_t sig[(RSA_NBITS + 7) / 8];
    size_t siglen = sizeof(sig);

    // Sign the digest
    if (ccrsa_sign_pkcs1v15_msg_blinded(&ccrng_zero, full_key, ccsha256_di(), MESSAGE_LEN, MESSAGE, &siglen, sig)) {
        failf("ccrsa_sign_pkcs1v15");
        return CCPOST_GENERIC_FAILURE;
    }

    // Verify computed signature == expected signature
    if (cc_cmp_safe(siglen, sig, SIGNATURE)) {
        failf("sig != expected signature");
        return CCPOST_KAT_FAILURE;
    }

    return 0;
}

CC_INLINE int fipspost_post_rsa_verify(uint32_t fips_mode)
{
    cc_size n = ccn_nof(RSA_NBITS);
    ccrsa_full_ctx_decl(ccn_sizeof(RSA_NBITS), full_key);
    ccrsa_ctx_n(full_key)=n;

    // Import the key
    if (ccrsa_import_priv(full_key, fipspost_post_rsa_test_key_nbytes, fipspost_post_rsa_test_key)) {
        failf("import");
        return CCPOST_GENERIC_FAILURE;
    }

    ccrsa_pub_ctx_t pub_key = ccrsa_ctx_public(full_key);

    // Verify Signature
    size_t siglen = (RSA_NBITS + 7) / 8;
    if (ccrsa_verify_pkcs1v15_msg(pub_key, ccsha256_di(), MESSAGE_LEN, MESSAGE, siglen, SIGNATURE, NULL)) {
        failf("ccrsa_verify_pkcs1v15");
        return CCPOST_KAT_FAILURE;
    }

    ccrsa_full_ctx_clear_nbits(RSA_NBITS, full_key);

    return 0;
}

// Test RSA Signature Creation and Verification
int fipspost_post_rsa_sig(uint32_t fips_mode)
{
    int ret_s = fipspost_post_rsa_sign(fips_mode);
    int ret_v = fipspost_post_rsa_verify(fips_mode);
    return ret_s | ret_v;
}
